@@ -1,6 +1,6 @@
 ;;;; chicken.scm - The CHICKEN Scheme compiler (loader/main-module)
 ;
-; Copyright (c) 2008-2010, The Chicken Team
+; Copyright (c) 2008-2013, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -28,7 +28,7 @@
 (declare
   (uses chicken-syntax chicken-ffi-syntax 
 	srfi-1 srfi-4 utils files extras data-structures support
-	compiler optimizer unboxing compiler-syntax scrutinizer driver platform backend 
+	compiler optimizer compiler-syntax scrutinizer driver platform backend 
 	srfi-69))
 
 
@@ -85,20 +85,43 @@
 		    (set! options (cons 'optimize-leaf-routines options)) )
 		   ((2)
 		    (set! options 
-		      (cons* 'optimize-leaf-routines 'inline options)) ) 
+		      (cons* 'optimize-leaf-routines
+			     'inline
+			     ;XXX 'clustering
+			     options)) ) 
 		   ((3)
 		    (set! options
-		      (cons* 'optimize-leaf-routines 'inline 'inline-global  options) ) )
+		      (cons* 'optimize-leaf-routines
+			     'inline
+			     'inline-global
+			     'local
+			     ;XXX 'clustering
+			     'specialize
+			     options) ) )
 		   ((4)
 		    (set! options
-		      (cons* 'optimize-leaf-routines 'inline 'local 'unboxing 'unsafe
+		      (cons* 'optimize-leaf-routines
+			     'inline 
+			     'inline-global
+			     'specialize
+			     ;XXX 'clustering
+			     'local 'unsafe
 			     options) ) )
 		   (else
 		    (when (>= level 5)
 		      (set! options 
-			(cons* 'disable-interrupts 'no-trace 'unsafe 'block
-			       'optimize-leaf-routines 'lambda-lift 'no-lambda-info
-			       'inline 'unboxing
+			(cons* 'disable-interrupts 
+			       'no-trace
+			       'unsafe
+			       'block
+			       'optimize-leaf-routines
+			       'no-lambda-info
+			       'specialize
+			       'optimize-leaf-routines 
+			       'no-lambda-info
+			       'inline
+			       'inline-global
+			       'clustering
 			       options) ) ) ) )
 		 (loop (cdr rest)) ) )
 	      ((eq? 'debug-level o)
@@ -115,7 +138,7 @@
 		   (quit "missing argument to `-~s' option" o) ) )
 	      (else
 	       (warning 
-		"invalid compiler option (gnored)" 
+		"invalid compiler option (ignored)" 
 		(if (string? o) o (conc "-" o)) )
 	       (loop rest) ) ) ) ) )
   (apply compile-source-file filename options)
