@@ -1,6 +1,6 @@
 ;;;; csc.scm - Driver program for the CHICKEN compiler - felix -*- Scheme -*-
 ;
-; Copyright (c) 2008-2010, The Chicken Team
+; Copyright (c) 2008-2013, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -45,6 +45,7 @@
 (define-foreign-variable INSTALL_MORE_STATIC_LIBS c-string "C_INSTALL_MORE_STATIC_LIBS")
 (define-foreign-variable INSTALL_SHARE_HOME c-string "C_INSTALL_SHARE_HOME")
 (define-foreign-variable INSTALL_LIB_HOME c-string "C_INSTALL_LIB_HOME")
+(define-foreign-variable INSTALL_LIB_NAME c-string "C_INSTALL_LIB_NAME")
 (define-foreign-variable INSTALL_INCLUDE_HOME c-string "C_INSTALL_INCLUDE_HOME")
 (define-foreign-variable INSTALL_STATIC_LIB_HOME c-string "C_INSTALL_STATIC_LIB_HOME")
 (define-foreign-variable TARGET_MORE_LIBS c-string "C_TARGET_MORE_LIBS")
@@ -52,6 +53,7 @@
 (define-foreign-variable TARGET_BIN_HOME c-string "C_TARGET_BIN_HOME")
 (define-foreign-variable TARGET_SHARE_HOME c-string "C_TARGET_SHARE_HOME")
 (define-foreign-variable TARGET_LIB_HOME c-string "C_TARGET_LIB_HOME")
+(define-foreign-variable TARGET_LIB_NAME c-string "C_TARGET_LIB_NAME")
 (define-foreign-variable TARGET_INCLUDE_HOME c-string "C_TARGET_INCLUDE_HOME")
 (define-foreign-variable TARGET_STATIC_LIB_HOME c-string "C_TARGET_STATIC_LIB_HOME")
 (define-foreign-variable TARGET_RUN_LIB_HOME c-string "C_TARGET_RUN_LIB_HOME")
@@ -64,10 +66,10 @@
 ;;; Parameters:
 
 (define mingw (eq? (build-platform) 'mingw32))
-(define msvc (eq? (build-platform) 'msvc))
 (define osx (eq? (software-version) 'macosx))
-(define win (or mingw msvc))
+(define win mingw)
 (define netbsd (eq? (software-version) 'netbsd))
+(define cygwin (eq? (build-platform) 'cygwin))
 
 (define elf
   (memq (software-version) '(linux netbsd freebsd solaris openbsd)))
@@ -103,24 +105,27 @@
 (define compiler (quotewrap (if host-mode INSTALL_CC TARGET_CC)))
 (define c++-compiler (quotewrap (if host-mode INSTALL_CXX TARGET_CXX)))
 (define rc-compiler (quotewrap (if host-mode INSTALL_RC_COMPILER TARGET_RC_COMPILER)))
-(define linker (quotewrap (if msvc "link" (if host-mode INSTALL_CC TARGET_CC))))
-(define c++-linker (quotewrap (if msvc "link" (if host-mode INSTALL_CXX TARGET_CXX))))
-(define object-extension (if msvc "obj" "o"))
-(define library-extension (if msvc "lib" "a"))
-(define link-output-flag (if msvc "-out:" "-o "))
-(define executable-extension (if msvc "exe" ""))
-(define compile-output-flag (if msvc "-Fo" "-o "))
+(define linker (quotewrap (if host-mode INSTALL_CC TARGET_CC)))
+(define c++-linker (quotewrap (if host-mode INSTALL_CXX TARGET_CXX)))
+(define object-extension "o")
+(define library-extension "a")
+(define link-output-flag "-o ")
+(define executable-extension "")
+(define compile-output-flag "-o ")
 (define nonstatic-compilation-options '())
 (define shared-library-extension ##sys#load-dynamic-extension)
 (define default-translation-optimization-options '())
-(define pic-options (if (or mingw msvc) '("-DPIC") '("-fPIC" "-DPIC")))
+(define pic-options (if (or mingw cygwin) '("-DPIC") '("-fPIC" "-DPIC")))
 (define windows-shell WINDOWS_SHELL)
 (define generate-manifest #f)
 
+(define libchicken
+  (if cygwin
+      (string-append "cyg" INSTALL_LIB_NAME "-0")
+      (string-append "lib" INSTALL_LIB_NAME)))
+
 (define default-library
-  (string-append
-   (if msvc "libchicken-static." "libchicken.")
-   library-extension))
+  (string-append libchicken "." library-extension))
 
 (define default-compilation-optimization-options (string-split (if host-mode INSTALL_CFLAGS TARGET_CFLAGS)))
 (define best-compilation-optimization-options default-compilation-optimization-options)
@@ -132,36 +137,40 @@
   '(-explicit-use -no-trace -no-warnings -no-usual-integrations -optimize-leaf-routines -unsafe
     -block -disable-interrupts -fixnum-arithmetic -to-stdout -profile -raw -accumulate-profile
     -check-syntax -case-insensitive -shared -compile-syntax -no-lambda-info
-    -lambda-lift -dynamic -disable-stack-overflow-checks -local
+    -dynamic -disable-stack-overflow-checks -local
     -emit-external-prototypes-first -inline -release -scrutinize
     -analyze-only -keep-shadowed-macros -inline-global -ignore-repository
     -no-symbol-escape -no-parentheses-synonyms -r5rs-syntax
     -no-argc-checks -no-bound-checks -no-procedure-checks -no-compiler-syntax
-    -emit-all-import-libraries -setup-mode -unboxing -no-elevation -no-module-registration
-    -no-procedure-checks-for-usual-bindings
+    -emit-all-import-libraries -setup-mode -no-elevation -no-module-registration
+    -no-procedure-checks-for-usual-bindings -module
+    -specialize -strict-types -clustering
+    -lambda-lift -unboxing		; OBSOLETE
     -no-procedure-checks-for-toplevel-bindings))
 
 (define-constant complex-options
-  '(-debug -output-file -heap-size -nursery -stack-size -compiler -unit -uses -keyword-style
+  '(-debug -heap-size -nursery -stack-size -compiler -unit -uses -keyword-style
     -optimize-level -include-path -database-size -extend -prelude -postlude -prologue -epilogue 
     -inline-limit -profile-name
-    -disable-warning			; OBSOLETE
-    -emit-inline-file -types
-    -feature -debug-level -heap-growth -heap-shrinkage -heap-initial-size -consult-inline-file
-    -emit-import-library -static-extension -no-feature))
+    -emit-inline-file -types -emit-type-file
+    -feature -debug-level 
+    -heap-growth -heap-shrinkage -heap-initial-size ; DEPRECATED
+    -consult-inline-file
+    -emit-import-library
+    -no-feature))
 
 (define-constant shortcuts
   '((-h "-help")
     (-s "-shared")
     (-S "-scrutinize")
+    (-M "-module")
     (|-P| "-check-syntax")
-    (|-V| "-version")
+    (|-V| "-version")			; DEPRECATED
     (-f "-fixnum-arithmetic")
     (|-D| "-feature")
     (-i "-case-insensitive")
     (|-K| "-keyword-style")
     (|-X| "-extend")
-    (|-N| "-no-usual-integrations")	; DEPRECATED
     (|-J| "-emit-all-import-libraries")
     (-x "-explicit-use")
     (-u "-unsafe")
@@ -170,7 +179,7 @@
     (-b "-block") ) )
 
 (define short-options
-  (string->list "PHhsfiENxubvwAOeWkctgSJ") )
+  (string->list "PHhsfiENxubvwAOeWkctgSJM") )
 
 
 ;;; Variables:
@@ -215,9 +224,7 @@
 	     (string-append "/" default-library)))) ))
 
 (define default-shared-library-files 
-  (if msvc
-      (list (string-append "libchicken." library-extension))
-      '("-lchicken")))
+  (list (string-append "-l" (if host-mode INSTALL_LIB_NAME TARGET_LIB_NAME))))
 
 (define library-files default-library-files)
 (define shared-library-files default-shared-library-files)
@@ -225,7 +232,8 @@
 (define translate-options '())
 
 (define include-dir
-  (let ((id (prefix "" "include" 
+  (let ((id (prefix ""
+                    (make-pathname "include" "chicken")
 		    (if host-mode INSTALL_INCLUDE_HOME TARGET_INCLUDE_HOME))))
     (and (not (member id '("/usr/include" "")))
 	 id) ) )
@@ -233,7 +241,12 @@
 (define compile-options '())
 
 (define builtin-compile-options
-  (if include-dir (list (conc "-I\"" include-dir "\"")) '()))
+  (append
+   (if include-dir (list (conc "-I\"" include-dir "\"")) '())
+   (cond ((get-environment-variable "CHICKEN_C_INCLUDE_PATH") => 
+	  (lambda (path) 
+	    (map (cut string-append "-I\"" <> "\"") (string-split path ":;"))))
+	 (else '()))))
 
 (define compile-only-flag "-c")
 (define translation-optimization-options default-translation-optimization-options)
@@ -249,19 +262,27 @@
 (define link-options '())
 
 (define (builtin-link-options)
-  (cond (elf
-	 (list
-	  (conc "-L\"" library-dir "\"")
-	  (conc " -Wl,-R\""
-		(if (and deployed (not netbsd))
-		    "\\$ORIGIN"
-		    (prefix "" "lib"
-			    (if host-mode
-				INSTALL_LIB_HOME
-				TARGET_RUN_LIB_HOME)))
-		"\"")) )
-	(else
-	 (list (conc "-L\"" library-dir "\"")))))
+  (append
+   (cond (elf
+	  (list
+	   (conc "-L\"" library-dir "\"")
+	   (conc " -Wl,-R\""
+		 (if (and deployed (not netbsd))
+		     "\\$ORIGIN"
+		     (prefix "" "lib"
+			     (if host-mode
+				 INSTALL_LIB_HOME
+				 TARGET_RUN_LIB_HOME)))
+		 "\"")) )
+	 (else
+	  (list (conc "-L\"" library-dir "\""))))
+   (if (and deployed (eq? (software-version) 'freebsd))
+       (list "-z origin")
+       '())
+   (cond ((get-environment-variable "CHICKEN_C_LIBRARY_PATH") => 
+	  (lambda (path) 
+	    (map (cut string-append "-L\"" <> "\"") (string-split path ":;"))))
+	 (else '()))))
 	
 (define target-filename #f)
 (define verbose #f)
@@ -272,7 +293,6 @@
 (define shared #f)
 (define static #f)
 (define static-libs #f)
-(define static-extensions '())
 (define required-extensions '())
 
 
@@ -293,12 +313,12 @@ Usage: #{csc} FILENAME | OPTION ...
   General options:
 
     -h  -help                      display this text and exit
-    -v                             show intermediate compilation stages
-    -vv  -verbose                  display information about translation
+    -v  -verbose                   show compiler notes and tool-invocations
+    -vv                            display information about translation
                                     progress
     -vvv                           display information about all compilation
                                     stages
-    -V  -version                   display Scheme compiler version and exit
+    -version                       display Scheme compiler version and exit
     -release                       display release number and exit
 
   File and pathname options:
@@ -320,11 +340,11 @@ Usage: #{csc} FILENAME | OPTION ...
   Syntax related options:
 
     -i -case-insensitive           don't preserve case of read symbols    
-    -K  -keyword-style STYLE       enable alternative keyword-syntax
+    -K -keyword-style STYLE        enable alternative keyword-syntax
                                     (prefix, suffix or none)
-        -no-parentheses-synonyms   disables list delimiter synonyms
-        -no-symbol-escape          disables support for escaped symbols
-        -r5rs-syntax               disables the Chicken extensions to
+       -no-parentheses-synonyms    disables list delimiter synonyms
+       -no-symbol-escape           disables support for escaped symbols
+       -r5rs-syntax                disables the Chicken extensions to
                                     R5RS syntax
     -compile-syntax                macros are made available at run-time
     -j -emit-import-library MODULE write compile-time module information into
@@ -332,6 +352,7 @@ Usage: #{csc} FILENAME | OPTION ...
     -J -emit-all-import-libraries  emit import-libraries for all defined modules
     -no-module-registration        do not generate module registration code
     -no-compiler-syntax            disable expansion of compiler-macros
+    -M -module                     wrap compiled code into implicit module
 
   Translation options:
 
@@ -366,15 +387,15 @@ Usage: #{csc} FILENAME | OPTION ...
     -b  -block                     enable block-compilation
     -disable-interrupts            disable interrupts in compiled code
     -f  -fixnum-arithmetic         assume all numbers are fixnums
-    -lambda-lift                   perform lambda-lifting
     -disable-stack-overflow-checks disables detection of stack-overflows
     -inline                        enable inlining
-    -inline-limit                  set inlining threshold
+    -inline-limit LIMIT            set inlining threshold
     -inline-global                 enable cross-module inlining
-    -unboxing                      use unboxed temporaries if possible
+    -specialize                    perform type-based specialization of primitive calls
     -n -emit-inline-file FILENAME  generate file with globally inlinable
                                     procedures (implies -inline -local)
     -consult-inline-file FILENAME  explicitly load inline file
+    -emit-type-file FILENAME       write type-declaration information into file
     -no-argc-checks                disable argument count checks
     -no-bound-checks               disable bound variable checks
     -no-procedure-checks           disable procedure call checks
@@ -384,15 +405,15 @@ Usage: #{csc} FILENAME | OPTION ...
     -no-procedure-checks-for-toplevel-bindings
                                    disable procedure call checks for toplevel
                                     bindings
+    -strict-types                  assume variable do not change their type
+    -clustering                    combine groups of local procedures into dispatch
+                                     loop
 
   Configuration options:
 
     -unit NAME                     compile file as a library unit
     -uses NAME                     declare library unit as used.
     -heap-size NUMBER              specifies heap-size of compiled executable
-    -heap-initial-size NUMBER      specifies heap-size at startup time
-    -heap-growth PERCENTAGE        specifies growth-rate of expanding heap
-    -heap-shrinkage PERCENTAGE     specifies shrink-rate of contracting heap
     -nursery NUMBER  -stack-size NUMBER
                                    specifies nursery size of compiled
                                    executable
@@ -431,8 +452,6 @@ Usage: #{csc} FILENAME | OPTION ...
     -static-libs                   link with static CHICKEN libraries
     -static                        generate completely statically linked
                                     executable
-    -static-extension NAME         link extension NAME statically
-                                    (if available)
     -F<DIR>                        pass \"-F<DIR>\" to C compiler
                                     (add framework header path on Mac OS X)
     -framework NAME                passed to linker on Mac OS X
@@ -465,8 +484,9 @@ Usage: #{csc} FILENAME | OPTION ...
     -host                          compile for host when configured for
                                     cross-compiling
     -private-repository            load extensions from executable path
-    -deployed                      compile support file to be used from a deployed 
-                                    executable
+    -deployed                      link support file to be used from a deployed 
+                                    executable (sets `rpath' accordingly, if supported
+                                    on this platform)
     -no-elevation                  embed manifest on Windows to supress elevation
                                     warnings for programs named `install' or `setup'
 
@@ -482,6 +502,7 @@ Usage: #{csc} FILENAME | OPTION ...
   every invocation of `#{csc}'.
 
 EOF
+;|        (for emacs font-lock)
   ) ) )
 
 
@@ -502,25 +523,25 @@ EOF
     (set! link-options
       (cons (cond
              (osx (if lib "-dynamiclib" "-bundle"))
-             (msvc "-dll")
              (else "-shared")) link-options))
     (set! shared #t) )
 
   (define (use-private-repository)
     (set! compile-options (cons "-DC_PRIVATE_REPOSITORY" compile-options))
     (when osx
+      ;; needed for C_path_to_executable (see chicken.h):
       (set! link-options (cons "-framework CoreFoundation" link-options))))
 
   (let loop ((args args))
     (cond [(null? args)
-	   ;; Builtin search directory options do not override explict options
+	   ;; Builtin search directory options do not override explicit options
            (set! compile-options (append compile-options builtin-compile-options))
            (set! link-options (append link-options (builtin-link-options)))
 	   ;;
 	   (when inquiry-only
 	     (when show-cflags (print* (compiler-options) #\space))
 	     (when show-ldflags (print* (linker-options) #\space))
-	     (when show-libs (print* (linker-libraries #t) #\space))
+	     (when show-libs (print* (linker-libraries) #\space))
 	     (newline)
 	     (exit) )
 	   (cond [(null? scheme-files)
@@ -597,14 +618,14 @@ EOF
 	       [(-libs)
 		(set! inquiry-only #t)
 		(set! show-libs #t) ]
-	       [(-v -verbose)
-		(when (and (number? verbose) (not msvc))
+	       ((-v -verbose)
+		(when (number? verbose)
 		  (set! compile-options (cons* "-v" "-Q" compile-options))
-		  (set! link-options (cons (if msvc "-VERBOSE" "-v") link-options)) )
-		(cond (verbose
-		       (t-options "-verbose") 
-		       (set! verbose 2)) 
-		      (else (set! verbose #t))) ]
+		  (set! link-options (cons "-v" link-options)) )
+		(t-options "-verbose")
+		(if verbose
+		    (set! verbose 2)
+		    (set! verbose #t)) )
 	       [(-w -no-warnings)
 		(set! compile-options (cons "-w" compile-options))
 		(t-options "-no-warnings") ]
@@ -625,33 +646,22 @@ EOF
 		(set! required-extensions (append required-extensions (list (car rest))))
 		(t-options "-require-extension" (car rest))
 		(set! rest (cdr rest)) ]
-	       [(-static-extension)
-		(check s rest)
-		(set! static-extensions (append static-extensions (list (car rest))))
-		(t-options "-static-extension" (car rest))
-		(set! rest (cdr rest)) ]
 	       ((-private-repository)
 		(use-private-repository))
 	       ((-no-elevation)
 		(set! generate-manifest #t))
-	       [(-gui
-		 -windows |-W|)		;DEPRECATED
+	       [(-gui)
 		(set! gui #t)
 		(set! compile-options (cons "-DC_GUI" compile-options))
-		(set! object-files 
-		  (cons (make-pathname 
-			 INSTALL_SHARE_HOME "chicken.rc"
-			 object-extension) 
-			object-files))
-		(when (or msvc mingw)
-		  (cond
-		   (mingw
-		    (set! link-options
-		      (cons* "-lkernel32" "-luser32" "-lgdi32" "-mwindows"
-			     link-options)))
-		   (msvc
-		    (set! link-options
-		      (cons* "kernel32.lib" "user32.lib" "gdi32.lib" link-options)))))]
+		(when mingw
+		  (set! object-files 
+		    (cons (make-pathname 
+			   INSTALL_SHARE_HOME "chicken.rc"
+			   object-extension) 
+			  object-files))
+		  (set! link-options
+		    (cons* "-lkernel32" "-luser32" "-lgdi32" "-mwindows"
+			   link-options)))]
 	       ((-deploy)
 		(set! deploy #t)
 		(set! deployed #t))
@@ -662,7 +672,7 @@ EOF
 		(when osx 
 		  (set! link-options (cons* "-framework" (car rest) link-options)) )
 		(set! rest (cdr rest)) ]
-	       [(-o)
+	       [(-o -output-file)
 		(check s rest)
 		(let ([fn (car rest)])
 		  (set! rest (cdr rest))
@@ -673,13 +683,10 @@ EOF
 	       [(|-O3|) (set! rest (cons* "-optimize-level" "3" rest))]
 	       [(|-O4|) (set! rest (cons* "-optimize-level" "4" rest))]
 	       [(|-O5|)
-		(set! rest (cons* "-optimize-level" "5" rest))
-		(when (memq (build-platform) '(mingw32 cygwin gnu clang))
-		  (set! compile-options 
-		    (cons* "-O3" "-fomit-frame-pointer" compile-options)) ) ]
-	       [(-d0) (set! rest (cons* "-debug-level" "0" rest))]
-	       [(-d1) (set! rest (cons* "-debug-level" "1" rest))]
-	       [(-d2) (set! rest (cons* "-debug-level" "2" rest))]
+		(set! rest (cons* "-optimize-level" "5" rest))]
+	       [(|-d0|) (set! rest (cons* "-debug-level" "0" rest))]
+	       [(|-d1|) (set! rest (cons* "-debug-level" "1" rest))]
+	       [(|-d2|) (set! rest (cons* "-debug-level" "2" rest))]
 	       [(-dry-run) 
 		(set! verbose #t)
 		(set! dry-run #t)]
@@ -720,13 +727,14 @@ EOF
 		(check s rest)
 		(set! rpath (car rest))
 		(when (and (memq (build-platform) '(gnu clang))
-			   (not mingw))
-		  (set! link-options (append link-options (list (string-append "-Wl,-R" rpath))))
-		  (set! rest (cdr rest)) ) ]
+			   (not mingw) (not osx))
+		  (set! link-options (append link-options (list (string-append "-Wl,-R" rpath)))) )
+	  	(set! rest (cdr rest)) ]
 	       [(-host) #f]
 	       [(-) 
-		(set! target-filename (make-pathname #f "a" executable-extension))
-		(set! scheme-files (append scheme-files '("-")))]
+		(set! scheme-files (append scheme-files '("-")))
+		(unless target-filename
+		  (set! target-filename (make-pathname #f "a" executable-extension)))]
 	       [else
 		(when (eq? s '-to-stdout) 
 		  (set! to-stdout #t)
@@ -738,8 +746,8 @@ EOF
 		      [(memq s simple-options) (t-options arg)]
 		      [(memq s complex-options) 
 		       (check s rest)
-		       (let* ([n (car rest)]
-			      [ns (string->number n)] )
+		       (let* ((n (car rest))
+			      (ns (string->number n)) )
 			 (t-options arg n)
 			 (set! rest (cdr rest)) ) ]
 		      [(and (> (string-length arg) 2) (string=? "-:" (substring arg 0 2)))
@@ -812,6 +820,9 @@ EOF
 		 (if to-stdout 
 		     '("-to-stdout")
 		     `("-output-file" ,(quotewrap fc)) )
+		 (if (##sys#fudge 13)
+		     '("-:d")
+		     '())
 		 (map quote-option
 		      (append 
 		       extra-features
@@ -827,7 +838,7 @@ EOF
   (unless keep-files (for-each $delete-file generated-scheme-files)) )
 
 
-;;; Compile all C  and .rc files:
+;;; Compile all C/C++  and .rc files:
 
 (define (run-compilation)
   (let ((ofiles '()))
@@ -841,6 +852,9 @@ EOF
 		 (quotewrap f)
 		 (string-append compile-output-flag (quotewrap fo)) 
 		 compile-only-flag
+		 (if (and cpp-mode (string=? "g++" c++-compiler))
+		     "-Wno-write-strings"
+		     "")
 		 (compiler-options) ) ) )
 	 (set! generated-object-files (cons fo generated-object-files))
 	 (set! ofiles (cons fo ofiles))))
@@ -876,9 +890,7 @@ EOF
 ;;; Link object files and libraries:
 
 (define (run-linking)
-  (let* ((files (map quotewrap
-		     (append object-files
-			     (nth-value 0 (static-extension-info)) ) ) )
+  (let* ((files (map quotewrap object-files))
 	 (target (quotewrap target-filename))
 	 (targetdir #f))
     (when deploy
@@ -906,13 +918,13 @@ EOF
 	      files
 	      (list (string-append link-output-flag (quotewrap target-filename))
 		    (linker-options)
-		    (linker-libraries #f) ) ) ) ) )
+		    (linker-libraries) ) ) ) ) )
     (when (and osx (or (not cross-chicken) host-mode))
       (command
        (string-append
-	"install_name_tool -change libchicken.dylib "
+	"install_name_tool -change " libchicken ".dylib "
 	(quotewrap 
-	 (let ((lib "libchicken.dylib"))
+	 (let ((lib (string-append libchicken ".dylib")))
 	   (if deployed
 	       (make-pathname "@executable_path" lib)
 	       (make-pathname
@@ -953,10 +965,12 @@ EOF
 (define (copy-libraries targetdir)
   (let ((lib (make-pathname
 	      (target-lib-path) 
-	      "libchicken"
+	      libchicken
 	      (cond (osx "dylib")
-		    (win "dll")
-		    (else "so")))))
+		    ((or win cygwin) "dll")
+		    (else (string-append
+                           "so."
+                           (number->string BINARY_VERSION)))))))
     (copy-files lib targetdir)))
 
 (define (copy-files from to)
@@ -968,33 +982,15 @@ EOF
      (quotewrap from)
      (quotewrap to))))
 
-(define (static-extension-info)
-  (let ((rpath (repository-path)))
-    (if (and rpath (pair? static-extensions))
-	(let loop ((exts static-extensions) (libs '()) (opts '()))
-	  (if (null? exts)
-	      (values (reverse libs) (reverse opts))
-	      (let ((info (extension-information (car exts))))
-		(if info
-		    (let ((a (assq 'static info)) 
-			  (o (assq 'static-options info)) )
-		      (loop (cdr exts) 
-			(if a (cons (make-pathname rpath (cadr a)) libs) libs)
-			(if o (cons (cadr o) opts) opts) ) ) 
-		    (loop (cdr exts) libs opts)) ) ) )
-	(values '() '()) ) ) )
-
 (define (linker-options)
   (string-append
    (string-intersperse
-    (append linking-optimization-options link-options
-	    (nth-value 1 (static-extension-info)) ) )
-   (if (and static (not mingw) (not msvc) (not osx)) " -static" "") ) )
+    (append linking-optimization-options link-options))
+   (if (and static (not mingw) (not osx)) " -static" "") ) )
 
-(define (linker-libraries #!optional staticexts)
+(define (linker-libraries)
   (string-intersperse
    (append
-    (if staticexts (nth-value 0 (static-extension-info)) '())
     (if (or static static-libs)
         library-files
         shared-library-files)
